module Main where

import Graphics.UI.WX
import Graphics.UI.WXCore 
import Directory
import HXML
import IO
import XMLParse
import Dialogs
import Filter

main :: IO ()
main
  = start gui
  
gui :: IO ()
gui = do {

     --Global variables
       fileXML<-varCreate [];
       varSet fileXML "";
       lastConditions<-varCreate [];
       varSet lastConditions "";
       condList<-varCreate [];
       varSet condList "";
       lastVars<-varCreate [];
       varSet lastVars [];
       varList<-varCreate [];
       varSet varList [];
       textTags<-varCreate [];
       varSet textTags [];
       lastTags<-varCreate [];
       varSet lastTags [];
       tagList<-varCreate [];
       varSet tagList [];
       nestQuery<-varCreate [];
       varSet nestQuery "";
       lastPattern<-varCreate [];
       varSet lastPattern [];
       patternTree<-varCreate [];
       varSet patternTree [];
       
     --Main Window
       mwin <- frame [text:="Phil 1.0"];
       
       mpan <- panel mwin [];
       nb <- notebook mpan [];
       build <- panel nb [];
       load <- panel nb [];
     
     --Load Panel
       loadq <- button load [text:="Load Query"];
       selectQuery <- singleListBox load [];
       execLoad <- button load [text:="Execute Query", tooltip:="Execute the selected query", enabled:=False];
       closeLoad <- button load [text:="Quit"];
       
     --Build Panel
       pinfo <- panel build [];
       ppatt <- panel build [];
       pcond <- panel build [];
       pfoot <- panel build [];
       
     --Info Build Panel
       loadXml <- button pinfo [text:="Load Xml", tooltip:="Load an XML-file"];
       selectFile <- singleListBox pinfo [];
       nesting <- button pinfo [text:="Add Nested Query", tooltip:="Select a query ..."];
       viewNesting <- textEntry pinfo [text:="No Nesting"];
       delNesting <- button pinfo [text:="Delete Nested Query"];
       textCtrlSetEditable viewNesting False;
       filterType <- singleListBox pinfo [items := ["Filter the best one","Filter all the best","Filter the best exact","Filter all exact"], selection:=0];
       pn <- radioBox pinfo Horizontal ["P","N"] [tooltip:="Filter Type\nP->Positive\nN->Negative"];
       resultInfo <- radioBox pinfo Vertical ["Embeddings","Number of embeddings"] [tooltip:="For the choice \"Number of embeddings\"\nyou have to specify the maximum cost"];
       costText <- staticText pinfo [text:="Maximum cost:"];
       cost <- textEntry pinfo [enabled:=False];
       
     --Pattern Build Panel
       addSons <- singleListBox ppatt [items:=["and", "or", "xor", "Single Node"],tooltip:="Select a choice to insert\nelements in the pattern", enabled:=False];
       viewPatt <- textCtrl ppatt [text:="", wrap:=WrapLine];
       textCtrlSetEditable viewPatt True;
       undo <- button ppatt [text:="Undo"];
       
     --Condition Build Panel
       addFunc <- comboBox pcond [items := ["Add Function", "match", "=", "<", "<=", ">", ">="], style:=wxCB_READONLY, selection:=0];
       undoCond <- button pcond [text:="Undo"];
       viewCond <- textCtrl pcond [text:="", wrap:=WrapLine];
       textCtrlSetEditable viewCond False;
     
     --Foot Build Panel
       discard <- button pfoot [text:="Discard Query"];
       save <- button pfoot [text:="Save Query", enabled:=False];
       execBuild <- button pfoot [text:="Execute Query"];
       closeBuild <- button pfoot [text:="Quit"];
       
     --Layout
       set mwin [layout:= container mpan $ tabs nb [
       				tab "Build" $
       					container build ( margin 5 $ fill $ column 5 [
       						container pinfo $ fill $ boxed "Generic Info" $
       							fill $ column 5 [
       								fill $ row 5 [vfill $ floatLeft $ widget loadXml, fill $ widget selectFile],
       								fill $ row 5 [vfill $ floatLeft $ widget nesting, hfill $ widget viewNesting, widget delNesting],
       								floatLeft $ fill $ row 5 [boxed "Query type" $ column 3 [widget filterType, widget pn], fill $ boxed "Result type" $ column 3 [widget resultInfo, fill $ row 3 [widget costText, hfill $ widget cost]]]
       							],
       						container ppatt $ fill $ boxed "Pattern" $
       							fill $ row 10 [column 5 [fill $ widget addSons, widget undo], fill $ widget viewPatt],
       						container pcond $ fill $ boxed "Conditions" $
       							fill $ row 10 [column 5 [vfill $ widget addFunc, widget undoCond], fill $ widget viewCond],
       						container pfoot $ fill $ row 10 [widget discard, widget save, widget execBuild, widget closeBuild]
       					]),
       				tab "Load" $
       					container load ( margin 5 $ fill $ column 5 [
       						floatLeft $ widget loadq,
       						floatLeft $ fill $ boxed "Queries" $ fill $ widget selectQuery,
       						floatLeft $ row 10 [widget execLoad, widget closeLoad]
       					])
       				]
       		];
       		
       windowSetClientSize mwin (sz 650 500);
       
     --Button Actions
       set closeBuild [on command := close mwin];
       
       set closeLoad [on command := close mwin];
       
       set loadXml [on command:= do {
			file<-fileOpenDialog mwin False True "XML File" [("Xml file (*.xml)",["*.xml"])] "" "";
			case file of
				Nothing -> return();
				Just filename -> itemAppend selectFile filename;
		  }];
       
       set loadq [on command:= do {
       			file<-fileOpenDialog mwin False True "Query File" [("Query file (*.phil)",["*.phil"])] "" "";
			case file of
				Nothing -> return();
				Just filename -> do {content<-readFile filename;
						     itemsDelete selectQuery;
						     extractQueryfromFile selectQuery content [];
						    }
       		}];
       
       set selectFile [on select := do {
       			i <- get selectFile selection;
			f <- get selectFile (item i);
			file <- return(modifyFileName f);
			varSet fileXML file;
       			names <- extractTags file;
       			varSet textTags names;
       			set addSons[enabled:=True];
       			set save[enabled:=True];
       		}];
       		
       set selectQuery [on select := set execLoad[enabled:=True]];
       
       set resultInfo [on select := do {
       			i <- get resultInfo selection;
			if (i==1)
			then set cost[enabled:=True]
			else set cost[enabled:=False]
       		}];
       
       set nesting [on command := do {
       			y<-nestingDialog mwin;
	       		case y of
				Just (query) -> do {
					set viewNesting[text:=query];
					varSet nestQuery query;
				}
				Nothing -> return();
       		}];
       	
       set delNesting [on command := do {
       			set viewNesting[text:="No Nesting"];
			varSet nestQuery "";
       		}];
       
       set discard [on command := do {
       			varSet patternTree [];
       			varSet nestQuery "";
       			varSet tagList [];
       			varSet varList [];
       			varSet condList [];
       			set viewPatt[text:=""];
       			set viewCond[text:=""];
       			set viewNesting[text:="No Nesting"];
       		}];
       
       set undo [on command := do {
       			pt <- varGet lastPattern;
       			tl <- varGet lastTags;
       			vl <- varGet lastVars;
       			varSet patternTree pt;
       			varSet tagList tl;
       			varSet varList vl;
       			viewTree <- return(showFormattedTerm 0 0 (head pt) ""); 
	       		set viewPatt[text:=viewTree];
       		}];
       
       set undoCond [on command := do {
       			cl <- varGet lastConditions;
       			varSet condList cl;
       			set viewCond [text:=cl];
       		}];
       
       set addSons [on select := do {
       			ptree <- varGet patternTree;
       			taglist <- varGet tagList;
       			if ((length ptree)/=0 && (length taglist)==0)
       			then warningDialog mwin "Attention" "No other insertions are allowed,\nthe pattern is full"
       			else do {
	       			i <- get addSons selection;
	       			son <- get addSons (item i);
	       			tTags <- varGet textTags;
	       			pTags <- return(controlTags(taglist));
	       			pVars <- varGet varList;
	       			y <- termDialog mwin tTags pTags pVars son;
	       			case y of
	       				Just (pos,newsons,newtags,newvars) -> do {
	       					ts<-varGet tagList;
	       					varSet lastTags ts;
	       					nts<-return(addnewTags ts pos newtags);
	       					varSet tagList nts;
	       					vs<-varGet varList;
	       					varSet lastVars vs;
	       					varSet varList (vs++newvars);
	       					varSet lastPattern ptree;
	       					nptree <- return(manageInsertion ptree son pos newsons newvars);
	       					varSet patternTree nptree;
	       					viewTree <- return(showFormattedTerm 0 0 (head nptree) ""); 
	       					-- putStrLn ("Actual tree: "++show(nptree));
	       					set viewPatt[text:=viewTree];
	       				}
	       				Nothing -> return();
	       		}
       		}];

       set addFunc [on select := do {
				i <- get addFunc selection;
				func <- get addFunc (item i);
				case i of
				0 -> return();
				1 -> do {
					set addFunc[selection:=0];
					y <- matchDialog mwin;
					case y of
						Just (arg1, arg2) -> do {
							conditions <- varGet condList;
			    				varSet lastConditions conditions;
			    				vars <- varGet varList;
			    				if (existsVar arg1 vars)
			    				then
			    					if (conditions == "")
			    					then do {varSet condList ("&"++arg1++" match "++arg2++"&");
								 	set viewCond [text:="&"++arg1++" match "++arg2++"&"];}
								else do {varSet condList (conditions++", "++"&"++arg1++" match "++arg2++"&");
								 	set viewCond [text:=conditions++", "++"&"++arg1++" match "++arg2++"&"];}
							else warningDialog mwin "Attention" ("The variable "++arg1++" does not occur in the pattern");
							}
						Nothing -> return(); }
				_ -> do {
					set addFunc[selection:=0];
					x <- funcDialog mwin func;
					case x of
						Just (arg1, arg2) -> do {
							conditions <- varGet condList;
			    				varSet lastConditions conditions;
			    				if (conditions == "")
			    				then do {varSet condList ("&"++arg1++" "++func++" "++arg2++"&");
								 set viewCond [text:="&"++arg1++" "++func++" "++arg2++"&"];}
							else do {varSet condList (conditions++", "++"&"++arg1++" "++func++" "++arg2++"&");
								 set viewCond [text:=conditions++", "++"&"++arg1++" "++func++" "++arg2++"&"];}
		    				}
						Nothing -> return();}
				}];
       
       set save [on command := do {
       			n <- get filterType selection;
       			i <- get pn selection;
       			posNeg <- get pn (item i);
       			i <- get resultInfo selection;
       			res <- get resultInfo (item i);
       			maxCost <- get cost text;
       			pattern <- varGet patternTree;
       			nes <- varGet nestQuery;
       			conditions <- varGet condList;
       			filename <- varGet fileXML;
       			if (null pattern)
       			then warningDialog mwin "Attention" "You have forgotten to specify the pattern"
       			else do {
       				pattern <- return(deleteBlanc(showFormattedTerm 0 0 (head pattern) ""));
	       			query <- return(composeQuery res maxCost n pattern nes filename conditions posNeg);
	       			fname<-fileSaveDialog mwin False False "Save Query" [("Phil Files", ["*.phil"])] "C:'\'XMLFiltering" "";
				case fname of
					Nothing->warningDialog mwin "Attention" "The path specified does not exist!"
					Just filepath -> appendFile filepath query
			}
       		}];
       
       set execBuild [on command := do {
       			n <- get filterType selection;
       			i <- get pn selection;
       			posNeg <- get pn (item i);
       			i <- get resultInfo selection;
       			res <- get resultInfo (item i);
       			maxCost <- get cost text;
       			pattern <- varGet patternTree;
       			nes <- varGet nestQuery;
       			conditions <- varGet condList;
       			filename <- varGet fileXML;
       			if (null pattern)
       			then warningDialog mwin "Attention" "You have forgotten to specify the pattern"
       			else do {
       				pattern <- return(deleteBlanc(showFormattedTerm 0 0 (head pattern) ""));
	       			query <- return(composeQuery res maxCost n pattern nes filename conditions posNeg);
	       			y <- filterXML2 query;
	       			case y of
	       				Just result -> do {
	       					m <- viewResultDialog mwin result;
	       					return();
	       				}
	       				Nothing -> return()
			}
       		}];
       
       set execLoad [on command := do {
       			i <- get selectQuery selection;
	    		query<-get selectQuery (item i);
	    		query<-return(tail(tail(tail(tail(tail(tail(tail(query))))))));
	    		y <- filterXML2 query;
	       		case y of
		       		Just result -> do {
		       			m <- viewResultDialog mwin result;
		       			return();
		       		}
		       		Nothing -> return()
       		}];
}

controlTags [] = ["Root"]
controlTags (x:xs) = (x:xs)

existsVar _ [] = False
existsVar s (l:ls)
	|s==l = True
	|otherwise = existsVar s ls

extractTags file = do{
	text<-readFile file;
	es<-return(parseDocument text);
	return(extTags es [])}

extTags [] tags = tags
extTags (x:xs) tags = case x of
	StartEvent gi atts -> extTags (attrs atts xs) (addTag gi tags)
	EndEvent _ -> extTags xs tags
	EmptyEvent gi atts -> extTags (attrs atts xs) (addTag gi tags)
	TextEvent s -> extTags xs tags
	
attrs [] es = es
attrs (x:xs) es = (StartEvent n []) : ((TextEvent s) : ((EndEvent n) : (attrs xs es)))
	where
	(n,s) = x

addTag::String->[String]->[String]
addTag lab [] = [lab]
addTag lab (t:ts)
	|lab==t = (t:ts)
	|lab<t = (lab:(t:ts))
	|lab>t = t:(addTag lab ts)

addnewTags [] _ ts = ts
addnewTags (x:xs) p ts
	|p==1 = (ts++xs)
	|otherwise = (x:(addnewTags xs (p-1) ts))

manageInsertion::[Tree XMLNode]->String->Int->[Tree XMLNode]->[String]->[Tree XMLNode]
manageInsertion [] "Single Node" _ t _ = t
manageInsertion [] b _ t _ = [(Tree (ELNode b []) t)]
manageInsertion [tree] "Single Node" pos tags vars = [(fst (addNode tree pos tags (addAttr vars)))]
manageInsertion [tree] b pos tags vars = [(fst (addNode tree pos [(Tree (ELNode b []) tags)] (addAttr vars)))]

addNode::Tree XMLNode->Int->[Tree XMLNode]->[(String,String)]->(Tree XMLNode,Int)
addNode (Tree (ELNode lab atts) (arg:args)) p sons natts = ((Tree (ELNode lab atts) nargs),np)
	where
	(nargs,np) = addNodeAux (arg:args) p sons natts
addNode (Tree (TXNode lab) []) p sons natts = ((Tree (TXNode lab) []),p)
addNode (Tree (ELNode lab atts) []) p sons natts
	|(p==1) && (notVarIn atts) = ((Tree (ELNode lab (atts++natts)) sons),0)
	|(p==1) = ((Tree (ELNode lab atts) []),p)
	|otherwise = ((Tree (ELNode lab atts) []),(p-1))

addNodeAux [] p sons atts = ([],p)
addNodeAux (t:ts) p sons atts = ((nt:nnt),nnp)
	where
	(nt,np) = addNode t p sons atts
	(nnt,nnp) = addNodeAux ts np sons atts

notVarIn [] = True
notVarIn (x:xs)
	|n=="var" = False
	|otherwise = notVarIn xs
	where
	(n,v)=x

addAttr [] = []
addAttr (v:vs) = ("var",v):(addAttr vs)

varFormat [] _ = []
varFormat (x:xs) []
	|n=="var" = (Tree (TXNode v) []):(varFormat xs [])
	|otherwise = varFormat xs []
	where
	(n,v)=x
varFormat (x:xs) [(Tree (ELNode lab []) args)] = [(Tree (ELNode lab []) ((varFormat (x:xs) [])++args))]

showAux prof [arg] _ i = showFormattedTerm prof i arg ""
showAux prof (a1:a2:args) node i
	|node=="and" = (showFormattedTerm prof i a1 " ,")++(showAux prof (a2:args) "and" i)
	|node=="or" = (showFormattedTerm prof i a1 " |")++(showAux prof (a2:args) "or" i)
	|node=="xor" = (showFormattedTerm prof i a1 " ?")++(showAux prof (a2:args) "xor" i)

showFormattedTerm prof i (Tree (ELNode name (a:as)) []) l = showFormattedTerm prof 1 (Tree (ELNode (controlName name (a:as)) []) (varFormat (a:as) [])) l 
showFormattedTerm prof i (Tree (ELNode name []) []) l = (concat(blancos (prof*sepBln)))++(name)++("()")++l++("\n")
showFormattedTerm prof i (Tree (TXNode text) []) l = (concat(blancos (prof*sepBln)))++(text)++l++("\n")
showFormattedTerm prof i (Tree (ELNode name []) [Tree (TXNode text) []]) l = if (((2*(length name))+(length text)+5+(prof*sepBln))>maxLengthElement) 
									 then (concat(blancos (prof*sepBln)))++(name)++("(\n")++(concat(blancos ((prof+1)*sepBln)))++text++("\n")++(concat(blancos (prof*sepBln)))++(")")++l++("\n")
									 else (concat(blancos (prof*sepBln)))++(name)++("(")++text++(")")++l++("\n")
showFormattedTerm prof i (Tree (ELNode name (a:as)) [arg]) l = showFormattedTerm prof 1 (Tree (ELNode (controlName name (a:as)) []) (varFormat (a:as) [arg])) l 
showFormattedTerm prof i (Tree (ELNode name []) [arg]) l = (concat(blancos (prof*sepBln)))++(name)++("(\n")++(showAux (prof+1) [arg] "" 1)++(concat(blancos (prof*sepBln)))++(")")++l++("\n")
showFormattedTerm prof i (Tree (ELNode name []) args) l 
	|i==0 = (concat(blancos (prof*sepBln)))++(name)++("(\n")++(showAux (prof+1) args "and" i)++(concat(blancos (prof*sepBln)))++(")")++l++("\n")
	|otherwise = showAux prof args name i

controlName n [] = n
controlName n (x:xs)
	|t=="type" = ('$':n)
	|otherwise = controlName n xs
	where
	(t,v)=x

composeQuery res maxCost n pattern nesting filename conditions posNeg
	|res == "Embeddings" = "\n"++(getFiltType n)++" "++pattern++" in "++rest1++rest2
	|maxCost=="" = "\ncount ("++(getFiltType n)++" "++pattern++" in "++rest1++rest2++")"
	|otherwise = "\ncount "++maxCost++" ("++(getFiltType n)++" "++pattern++" in "++rest1++rest2++")"
	where 
	rest1 = composeExp nesting filename
	rest2 = composeCond conditions posNeg
	composeExp nesting filename
		|nesting=="" = "file '"++filename++"' "
		|otherwise = "("++nesting++") "
	composeCond conditions posNeg
		|conditions=="" = "("++posNeg++")"
		|otherwise = "where "++conditions++" ("++posNeg++")"
	getFiltType i
		|i==0 = "filterOneBest"
		|i==1 = "filterAllBest"
		|i==2 = "filterBestExact"
		|i==3 = "filterAllExact"

deleteBlanc [] = []
deleteBlanc (x:xs) 
	| x=='\n' = deleteBlanc xs
	| x=='\t' = deleteBlanc xs
	| x==' ' = deleteBlanc xs
	|otherwise = x : deleteBlanc xs

modifyFileName [] = []
modifyFileName (x:xs)
	|x=='\\' = x:(x:(modifyFileName xs))
	|otherwise = x:(modifyFileName xs)